import os
import anndata
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from sklearn.preprocessing import MinMaxScaler
#### NEEDED FILES
# 1. GeneLength.txt


def counts2FPKM(counts,genelen):
    genelen = pd.read_csv(genelen, sep=',')
    genelen['TranscriptLength'] = genelen['Transcript end (bp)'] - genelen['Transcript start (bp)']
    genelen = genelen[['Gene name', 'TranscriptLength']]
    genelen = genelen.groupby('Gene name').max()
    # intersection
    inter = counts.columns.intersection(genelen.index)
    samplename = counts.index
    counts = counts[inter].values
    genelen = genelen.loc[inter].T.values
    # transformation
    totalreads = counts.sum(axis=1)
    counts = counts*1e9/(genelen*totalreads.reshape(-1,1))
    counts = pd.DataFrame(counts,columns=inter,index=samplename)
    return counts

def FPKM2TPM(fpkm):
    genename = fpkm.columns
    samplename = fpkm.index
    fpkm = fpkm.values
    total = fpkm.sum(axis=1).reshape(-1,1)
    fpkm = fpkm*1e6/total
    fpkm = pd.DataFrame(fpkm,columns=genename,index=samplename)
    return fpkm

def counts2TPM(counts,genelen):
    fpkm = counts2FPKM(counts,genelen)
    tpm = FPKM2TPM(fpkm)
    return tpm


def ProcessInputData(train_data, test_data, sep=None, datatype='TPM', variance_threshold=0.5,
                     genelenfile=None):

    ### read train data
    print('Reading training data')
    if type(train_data) is anndata.AnnData:
        pass
    elif type(train_data) is str:
        train_data = anndata.read_h5ad(train_data)
    train_data.var_names_make_unique()
    train_x = pd.DataFrame(train_data.X, columns=train_data.var.index)
    train_y = train_data.obs
    print('Reading is done')
    ### read test data
    print('Reading test data')
    if type(test_data) is str:
        test_x = pd.read_csv(test_data, index_col=0, sep=sep)
    elif type(test_data) is pd.DataFrame:
        test_x = test_data
    print('Reading test data is done')
    ### transform to datatype
    if datatype == 'FPKM':
        if genelenfile is None:
            raise Exception("Please add gene length file!")
        print('Transforming to FPKM')
        train_x = counts2FPKM(train_x,genelenfile)
    elif datatype == 'TPM':
        if genelenfile is None:
            raise Exception("Please add gene length file!")
        print('Transforming to TPM')
        train_x = counts2TPM(train_x,genelenfile)
    elif datatype == 'counts':
        print('Using counts data to train model')
    ### variance cutoff
    print('Variance Cutoff')
    test_x = test_x.loc[:, test_x.var(axis=0) > variance_threshold]

    ### find intersected genes
    print('Find intersected genes')
    inter = train_x.columns.intersection(test_x.columns)
    train_x = train_x[inter]
    test_x = test_x[inter]
    genename = list(inter)
    celltypes = train_y.columns
    samplename = test_x.index
    
    print('Intersected gene number is ',len(inter))
    ### MinMax process
    print('Log2 & MinMax scale')
    train_x = np.log2(train_x.values + 1)
    test_x = np.log2(test_x.values + 1)
    mms = MinMaxScaler()
    train_x = mms.fit_transform(train_x.T).T
    test_x = mms.fit_transform(test_x.T).T

    return train_x, train_y.values, test_x, genename, celltypes, samplename


def L1error(pred, true):
    return np.mean(np.abs(pred - true))

def CCCscore(y_pred, y_true, mode='all'):
    # pred: shape{n sample, m cell}
    if mode == 'all':
        y_pred = y_pred.reshape(-1,1)
        y_true = y_true.reshape(-1,1)
    elif mode == 'avg':
        pass
    ccc_value = 0
    for i in range(y_pred.shape[1]):
        r = np.corrcoef(y_pred[:, i], y_true[:, i])[0, 1]
        # Mean
        mean_true = np.mean(y_true[:, i])
        mean_pred = np.mean(y_pred[:, i])
        # Variance
        var_true = np.var(y_true[:, i])
        var_pred = np.var(y_pred[:, i])
        # Standard deviation
        sd_true = np.std(y_true[:, i])
        sd_pred = np.std(y_pred[:, i])
        # Calculate CCC
        numerator = 2 * r * sd_true * sd_pred
        denominator = var_true + var_pred + (mean_true - mean_pred) ** 2
        ccc = numerator / denominator
        ccc_value += ccc
    return ccc_value / y_pred.shape[1]

def score(pred, label):
    print('L1 error is', L1error(pred,label))
    print('CCC is ', CCCscore(pred, label))

def showloss(loss):
    sns.set()
    plt.plot(loss)
    plt.xlabel('iteration')
    plt.ylabel('loss')
    plt.show()






import torch
import numpy as np
import pandas as pd
from tqdm import tqdm
from torch.optim import Adam
import torch.nn.functional as F
from torch.utils.data import DataLoader

from .model import reproducibility, simdatset, AutoEncoder, device
from .utils import showloss

def training_stage(model, train_loader, optimizer, epochs=10):
    model.train()
    model.state = 'train'
    loss = []
    recon_loss = []
    for i in tqdm(range(epochs)):
        for k, (data, label) in enumerate(train_loader):
            optimizer.zero_grad()
            x_recon, cell_prop, sigm = model(data)
            batch_loss = F.l1_loss(cell_prop, label) + F.l1_loss(x_recon, data)
            batch_loss.backward()
            optimizer.step()
            loss.append(F.l1_loss(cell_prop, label).cpu().detach().numpy())
            recon_loss.append(F.l1_loss(x_recon, data).cpu().detach().numpy())

    return model, loss, recon_loss

def adaptive_stage(model, data, optimizerD, optimizerE, step=10, max_iter=5):
    data = torch.from_numpy(data).float().to(device)
    loss = []
    model.eval()
    model.state = 'test'
    _, ori_pred, ori_sigm = model(data)
    ori_sigm = ori_sigm.detach()
    ori_pred = ori_pred.detach()
    model.state = 'train'
    for k in range(max_iter):
        model.train()
        for i in range(step):
            optimizerD.zero_grad()
            x_recon, _, sigm = model(data)
            batch_loss = F.l1_loss(x_recon, data)+F.l1_loss(ori_sigm,sigm)
            batch_loss.backward()
            optimizerD.step()
            loss.append(F.l1_loss(x_recon, data).cpu().detach().numpy())

        for i in range(step):
            optimizerE.zero_grad()
            x_recon, pred, _ = model(data)
            batch_loss = F.l1_loss(x_recon, data) + F.l1_loss(ori_pred, pred)
            batch_loss.backward()
            optimizerE.step()
            loss.append(F.l1_loss(x_recon, data).cpu().detach().numpy())

    model.eval()
    model.state = 'test'
    _, pred, sigm = model(data)
    return sigm.cpu().detach().numpy(), loss, ori_pred.cpu().numpy()

def train_model(train_x, train_y,
                model_name=None,
                batch_size=128, iteration=10000):

    
    reproducibility(9)
    
    train_loader = DataLoader(simdatset(train_x, train_y), batch_size=batch_size, shuffle=True)
    model = AutoEncoder(train_x.shape[1], train_y.shape[1]).to(device)
    optimizer = Adam(model.parameters(), lr=1e-4)
    print('Start training')
    model, loss, reconloss = training_stage(model, train_loader, optimizer, epochs=int(iteration /(len(train_x)/128)))
    print('Training is done')
    print('prediction loss is:')
    showloss(loss)
    print('reconstruction loss is:')
    showloss(reconloss)
    if model_name is not None:
        print('Model is saved')
        torch.save(model, model_name+".pth")
    return model

def predict(test_x, genename, celltypes, samplename,
            model_name=None, model=None,
            adaptive=True, mode='overall'):
    reproducibility(9)
    
    if model is not None and model_name is None:
        print('Model is saved without defined name')
        torch.save(model, 'model.pth')
    if adaptive is True:
        if mode == 'high-resolution':
            TestSigmList = np.zeros((test_x.shape[0], len(celltypes), len(genename)))
            TestPred = np.zeros((test_x.shape[0], len(celltypes)))
            print('Start adaptive training at high-resolution')
            for i in tqdm(range(len(test_x))):
                x = test_x[i,:].reshape(1,-1)
                if model_name is not None and model is None:
                    model = torch.load(model_name + ".pth")
                elif model is not None and model_name is None:
                    model = torch.load("model.pth")
                decoder_parameters = [{'params': [p for n, p in model.named_parameters() if 'decoder' in n]}]
                encoder_parameters = [{'params': [p for n, p in model.named_parameters() if 'encoder' in n]}]
                optimizerD = torch.optim.Adam(decoder_parameters, lr=1e-4)
                optimizerE = torch.optim.Adam(encoder_parameters, lr=1e-4)
                test_sigm, loss, test_pred = adaptive_stage(model, x, optimizerD, optimizerE, step=500, max_iter=3)
                TestSigmList[i, :, :] = test_sigm
                TestPred[i,:] = test_pred
            TestPred = pd.DataFrame(TestPred,columns=celltypes,index=samplename)
            CellTypeSigm = {}
            for i in range(len(celltypes)):
                cellname = celltypes[i]
                sigm = TestSigmList[:,i,:]
                sigm = pd.DataFrame(sigm,columns=genename,index=samplename)
                CellTypeSigm[cellname] = sigm
            print('Adaptive stage is done')

            return CellTypeSigm, TestPred

        elif mode == 'overall':
            if model_name is not None and model is None:
                model = torch.load(model_name + ".pth")
            elif model is not None and model_name is None:
                model = torch.load("model.pth")
            decoder_parameters = [{'params': [p for n, p in model.named_parameters() if 'decoder' in n]}]
            encoder_parameters = [{'params': [p for n, p in model.named_parameters() if 'encoder' in n]}]
            optimizerD = torch.optim.Adam(decoder_parameters, lr=1e-4)
            optimizerE = torch.optim.Adam(encoder_parameters, lr=1e-4)
            print('Start adaptive training for all the samples')
            test_sigm, loss, test_pred = adaptive_stage(model, test_x, optimizerD, optimizerE, step=500, max_iter=3)
            print('Adaptive stage is done')
            test_sigm = pd.DataFrame(test_sigm,columns=genename,index=celltypes)
            test_pred = pd.DataFrame(test_pred,columns=celltypes,index=samplename)

            return test_sigm, test_pred

    else:
        if model_name is not None and model is None:
            model = torch.load(model_name+".pth")
        elif model is not None and model_name is None:
            model = model
        print('Predict cell fractions without adaptive training')
        model.eval()
        data = torch.from_numpy(test_x).float().to(device)
        _, pred, _ = model(data)
        pred = pred.cpu().detach().numpy()
        pred = pd.DataFrame(pred, columns=celltypes, index=samplename)
        print('Prediction is done')
        return pred






import pandas as pd
from .simulation import generate_simulated_data
from .utils import ProcessInputData
from .train import train_model, predict
from .model import scaden

def Deconvolution(sc_reference, real_bulk, sep='\t',
                  datatype='counts', genelenfile=None,
                  mode='overall', adaptive=True,
                  save_model_name=None):
    """
    :param sc_reference: a txt expression file path index is cell type name, columns is gene name
    :param real_bulk: an expression file path, index is sample, columns is gene name
    :param sep: used to read bulk data, depends on the format
    :param bulkDataType: FPKM or TPM, if type is RPKM, please just use FPKM
    :param genelenfile: specify the location of gene length file for transforming counts data to TPM or FPKM
                        this file should in txt format and
                            contain three columns: [Gene name,Transcript start (bp),Transcript end (bp)]
    :param mode: 'high-resolution' means this will apply adaptive stage to every single sample to generate signature matrix,
                 'overall' means that it will deconvolve all the samples at the same time
    :param adaptive: it has to be True, if model is 'high-resolution'
    :param save_model_name: the name used to save model, if it was not provided, it would not be saved
    :return: depends on the mode or adaptive
             there are three combinations:
             1. high-resolution and adaptive deconvolution
                this will return a dictionary and predicted cell fractions in pandas dataframe format
                the keys of the dict are the pre-defined cell type names in the single cell reference data
                the values of the dict are the dataframe of gene expression and samples
             2. overall and adaptive deconvolution
                this will return a signature matrix and a cell fraction
                the rows of the signature matrix is the gene expression in each cell types
                both of the variables are in dataframe format
             3. overall and non-adaptive deconvolution
                this will return a cell fraction directly
                the signature matrix in this mode is None
    """
    simudata = generate_simulated_data(sc_data=sc_reference, samplenum=5000)
    train_x, train_y, test_x, genename, celltypes, samplename = \
        ProcessInputData(simudata, real_bulk, sep=sep, datatype=datatype, genelenfile=genelenfile)
    print('training data shape is ', train_x.shape, '\ntest data shape is ', test_x.shape)
    if save_model_name is not None:
        model = train_model(train_x, train_y, save_model_name, batch_size=128, iteration=5000)
    else:
        model = train_model(train_x, train_y, batch_size=128, iteration=5000)

    print('Notice that you are using parameters: mode=' + str(mode) + ' and adaptive=' + str(adaptive))
    if adaptive is True:
        if mode == 'high-resolution':

            CellTypeSigm, Pred = \
                predict(test_x=test_x, genename=genename, celltypes=celltypes, samplename=samplename,
                        model=model, model_name=save_model_name,
                        adaptive=adaptive, mode=mode)
            return CellTypeSigm, Pred
        elif mode == 'overall':
            Sigm, Pred = \
                predict(test_x=test_x, genename=genename, celltypes=celltypes, samplename=samplename,
                        model=model, model_name=save_model_name,
                        adaptive=adaptive, mode=mode)
            return Sigm, Pred
    else:
        Pred = predict(test_x=test_x, genename=genename, celltypes=celltypes, samplename=samplename,
                       model=model, model_name=save_model_name,
                       adaptive=adaptive, mode=mode)
        Sigm = None
        return Sigm, Pred

def ScadenDeconvolution(sc_reference, real_bulk, sep='\t', genelenfile=None):
    simudata = generate_simulated_data(sc_data=sc_reference, samplenum=5000)
    train_x, train_y, test_x, genename, celltypes, samplename = \
        ProcessInputData(simudata, real_bulk, sep=sep, datatype='counts', genelenfile=genelenfile)
    print('training data shape is ', train_x.shape, '\ntest data shape is ', test_x.shape)
    pred = test_scaden(train_x,train_y,test_x,batch_size=128)
    pred = pd.DataFrame(pred, columns=celltypes, index=samplename)
    return pred

def test_scaden(train_x,train_y,test_x,batch_size=128):
    architectures = {'m256': ([256,128,64,32],[0,0,0,0]),
                     'm512': ([512,256,128,64],[0, 0.3, 0.2, 0.1]),
                     'm1024': ([1024, 512, 256, 128],[0, 0.6, 0.3, 0.1])}
    model = scaden(architectures, train_x, train_y, epochs=int(5000/(len(train_x)/batch_size)))
    model.train()
    pred = model.predict(test_x)
    return pred





import anndata
import numpy as np
import pandas as pd
from tqdm import tqdm
from numpy.random import choice

def generate_simulated_data(sc_data, outname=None,
                            prop = None,
                            n=500, samplenum=4000):
    # sc_data should be a cell*gene matrix, no null value, txt file, sep='\t'
    # index should be cell names
    # columns should be gene labels
    print('Reading single-cell dataset, this may take 1 min')
    sc_data = pd.read_csv(sc_data,index_col=0,sep='\t')
    print('Reading dataset is done')
    sc_data.dropna(inplace=True)
    sc_data['celltype'] = sc_data.index
    sc_data.index = range(len(sc_data))

    num_celltype = len(sc_data['celltype'].value_counts())
    genename = sc_data.columns[:-1]
    celltype_groups = sc_data.groupby('celltype').groups
    sc_data.drop(columns='celltype' ,inplace=True)

    # use ndarray to accelerate
    # change to C_CONTIGUOUS, 10x speed up
    sc_data = sc_data.values
    sc_data = np.ascontiguousarray(sc_data, dtype=np.float32)
    # make random cell proportions
    if prop is None:
        print('Generating random cell fractions for each pseudo-bulk sample')
        prop = np.random.rand(samplenum ,num_celltype)
        prop = prop /np.sum(prop ,axis=1).reshape(-1 ,1)

        for i in range(prop.shape[0]):
            indices = np.random.choice(np.arange(prop.shape[1]), replace=False, size=int(prop.shape[1] * 0.3))
            prop[i ,indices] = 0
        prop = prop /np.sum(prop ,axis=1).reshape(-1 ,1)
        print('Generating cell frations is done')

    # make the dictionary
    for key, value in celltype_groups.items():
        celltype_groups[key] = np.array(value)
    # precise number for each celltype
    cell_num = np.floor(n*prop)

    # start sampling
    sample = np.zeros((prop.shape[0] ,sc_data.shape[1]))
    allcellname = celltype_groups.keys()
    print('Sampling cells to compose pseudo-bulk data')
    for i ,sample_prop in tqdm(enumerate(cell_num)):
        for j, cellname in enumerate(allcellname):
            select_index = choice(celltype_groups[cellname] ,size=int(sample_prop[j]) ,replace=True)
            sample[i] += sc_data[select_index].sum(axis=0)
            
    prop = pd.DataFrame(prop ,columns=celltype_groups.keys())
    simulated_dataset = anndata.AnnData(X=sample,
                                        obs=prop,
                                        var=pd.DataFrame(index=genename))
    print('Sampling is done')
    if outname is not None:
        simulated_dataset.write_h5ad(outname+'.h5ad')
    return simulated_dataset